@@ -3,7 +3,8 @@ A module for identifying and representing different tracker formats, and the
 samples inside them.
 """
 from enum import Enum
-from io import SEEK_CUR
+from io import SEEK_CUR, BytesIO
+
 from . import pcm
 
 class LoopType(Enum):
@@ -22,7 +23,7 @@ def identify_module(file) -> str:
     if magic[:8] == b"ziRCONia":
         raise NotImplementedError("MMCMP-compression isn't supported.")
     if magic[:4] == b"\xc1\x83*\x9e":
-        raise NotImplementedError("UMX files aren't supported yet.")
+        return UnrealEngineUMX(file)
 
     file.seek(28)
     sig_one = file.read(2)
@@ -70,7 +71,7 @@ class ProtrackerMOD:
                 sample["rate"] = self.SAMPLE_RATE
                 sample["width"] = self.SAMPLE_WIDTH
                 sample["data"] = self.file.read(sample["length"])
-                sample["data"] = pcm.signed_to_unsigned(sample["data"])
+                sample["data"] = pcm.signed_to_unsigned_8bit(sample["data"])
 
     def get_sample_count(self) -> int:
         """Returns the # of samples present."""
@@ -132,8 +133,6 @@ class ProtrackerMOD:
 class ScreamTracker3S3M:
     """Retrieves sample data from ScreamTracker 3 S3M files."""
 
-    SAMPLE_WIDTH = 8 // 8
-
     def __init__(self, file):
         self.file = file
 
@@ -178,7 +177,7 @@ class ScreamTracker3S3M:
             if sample["length"] > 0:
                 sample["data"] = self.file.read(sample["length"])
                 if signed:
-                    sample["data"] = pcm.signed_to_unsigned(sample["data"])
+                    sample["data"] = pcm.signed_to_unsigned_8bit(sample["data"])
 
     @staticmethod
     def decode_sample_header(header_bytes) -> dict:
@@ -266,9 +265,12 @@ class ImpulseTrackerIT:
                     sample["data"] = sample_data
                 else:
                     sample["data"] = self.decompress_it_sample(sample_data)
-                # HACK: don't know why 8-bit samples are signed while 16-bit samples are unsigned
-                if sample["width"] == 1 or not sample["signed"]:
-                    sample["data"] = pcm.signed_to_unsigned(sample["data"])
+                # python's wave module always outputs 8-bit samples as unsigned,
+                # and 16-bit samples as signed.
+                if sample["signed"] and sample["width"] == 1:
+                    sample["data"] = pcm.signed_to_unsigned_8bit(sample["data"])
+                elif not sample["signed"] and sample["width"] == 2:
+                    raise NotImplementedError("Unsigned 16-bit samples aren't supported yet.")
 
     @staticmethod
     def decode_sample_header(header_bytes) -> dict:
@@ -281,14 +283,16 @@ class ImpulseTrackerIT:
         #skip DOS filename, blank and global volume
 
         flags = int.from_bytes(header_bytes[18:19], "big")
+
         # assert (flags >> 0) & 1 == 1
         # on = 16-bit, off = 8-bit
-        sample["width"] = 16//8 if (flags >> 1) & 1 else 8//8
-        if (flags >> 2) & 1:
+        sample["width"] = 16//8 if bool(flags & 0b00000010) else 8//8
+        stereo_flag = bool(flags & 0b00000100)
+        if stereo_flag:
             raise NotImplementedError("Stereo samples aren't supported.")
-        sample["compressed"] = bool((flags >> 3) & 1)
-        loop_flag = bool((flags >> 4) & 1)
-        ping_pong_flag = bool((flags >> 6) & 1)
+        sample["compressed"] = bool(flags & 0b00001000)
+        loop_flag = bool(flags & 0b00010000)
+        ping_pong_flag = bool(flags & 0b00100000)
         if loop_flag:
             if ping_pong_flag:
                 sample["loop_type"] = LoopType.PING_PONG
@@ -303,8 +307,8 @@ class ImpulseTrackerIT:
 
         # skip default pan
 
-        convert = int.from_bytes(header_bytes[46:47], "big")
-        sample["signed"] = bool(convert & 1)
+        convert = int.from_bytes(header_bytes[46:47], "little")
+        sample["signed"] = bool(convert & 0b00000001)
 
         # length of sample is stored in no. of samples NOT no. of bytes
         sample["length"] = int.from_bytes(header_bytes[48:52], "little") * sample["width"]
@@ -321,5 +325,100 @@ class ImpulseTrackerIT:
 
     @staticmethod
     def decompress_it_sample(sample_bytes) -> bytes:
-        # TODO: investigate compression flag
-        return sample_bytes
\ No newline at end of file
+        raise NotImplementedError("IT214 sample compression isn't supported yet.")
+
+class UnrealEngineUMX:
+    """Retrieves module file contained within an Unreal Engine UMX package file."""
+
+    def __init__(self, file):
+        self.file = file
+
+        self.file.seek(4)
+        version = int.from_bytes(file.read(4), "little")
+        if version < 61:
+            raise NotImplementedError("UMX files under version 61 aren't supported.")
+        self.file.seek(4, SEEK_CUR) # skip package flags
+
+        name_table_count = int.from_bytes(self.file.read(4), "little")
+        name_table_offset = int.from_bytes(self.file.read(4), "little")
+        export_table_count = int.from_bytes(self.file.read(4), "little")
+        export_table_offset = int.from_bytes(self.file.read(4), "little")
+
+        if export_table_count > 1:
+            raise TypeError("Unreal Package contains more than one exported object.")
+
+        self.file.seek(name_table_offset)
+        name_table_index = 0
+        names = []
+        while name_table_index < name_table_count:
+            if version > 61:
+                current_name_length = int.from_bytes(self.file.read(1), "little")
+                names.append(self.file.read(current_name_length - 1).decode("ascii"))
+                self.file.seek(1, SEEK_CUR) # skip terminating zero
+            else:
+                at_terminating_zero = False
+                name = str()
+                while not at_terminating_zero:
+                    byte = self.file.read(1)
+                    if not byte == b"\x00":
+                        name += byte.decode()
+                    else:
+                        names.append(name)
+                        at_terminating_zero = True
+            self.file.seek(4, SEEK_CUR) # skip  object flags
+            name_table_index += 1
+
+        if "Music" not in names:
+            raise TypeError("Unreal Package File does not contain music.")
+
+        self.file.seek(export_table_offset)
+        self.read_compact_index() # skip class index
+        self.read_compact_index() # skip super index
+        self.file.seek(4, SEEK_CUR) # skip package index
+        self.read_compact_index() # skip object name
+        self.file.seek(4, SEEK_CUR) # skip object flags
+        self.read_compact_index() # skip serial size
+        serial_offset = self.read_compact_index()
+
+        self.file.seek(serial_offset)
+        self.file.seek(2, SEEK_CUR) # skip chunk count
+        if version > 61:
+            self.file.seek(4, SEEK_CUR) # skip following byte position
+        chunk_size = self.read_compact_index() # serial size minus the object's header
+        embedded_stream = BytesIO(self.file.read(chunk_size))
+        embedded_file = identify_module(embedded_stream)
+
+        self.title = embedded_file.title
+        self.samples = embedded_file.samples
+
+    def read_compact_index(self):
+        """
+        Reads a byte (or more depending on continue flags) at self.file's stream
+        position.
+        Converts these byte(s) from a compact index into an integer, and returns
+        it.
+        """
+        compact_index = ""
+
+        first_byte = int.from_bytes(self.file.read(1), "little")
+        if first_byte == 0:
+            return 0
+        negative_flag = bool(first_byte & 0b10000000)
+        continue_flag = bool(first_byte & 0b01000000)
+        first_byte_binary = bin(first_byte & 0b00111111)[2:]
+        while len(first_byte_binary) < 6:
+            first_byte_binary = str(0) + first_byte_binary
+        compact_index = first_byte_binary
+
+        while continue_flag:
+            other_byte = int.from_bytes(self.file.read(1), "little")
+            other_byte_binary = bin(other_byte & 0b01111111)[2:]
+            while len(other_byte_binary) < 7:
+                other_byte_binary = str(0) + other_byte_binary
+            compact_index = other_byte_binary + compact_index
+            continue_flag = bool(other_byte & 0b10000000)
+
+        value = int(compact_index, 2)
+        if negative_flag:
+            value = -value
+        return value
